#include <stdlib.h>
#include <stdio.h>
#include <pthread.h>
#include <unistd.h>
#include <semaphore.h>
#include <time.h>

#define RAND_DIVISOR 5
#define BUFFER_SIZE 1
typedef int buffer_item;

pthread_mutex_t mutex;
sem_t full, empty;
buffer_item buffer[BUFFER_SIZE];
int counter;
int nLoops;
int numProd;
int numCons;

void *producer(void*);
void *consumer(void*);
int insert_item(buffer_item);
int remove_item(buffer_item*);

void initializeData() {
	pthread_mutex_init(&mutex, NULL);
	sem_init(&full, 0, 0);
	sem_init(&empty, 0, BUFFER_SIZE);
	counter = 0;
}

void *producer(void *param) {
	buffer_item item;
	int num = *((int*)param);
	for(int i = 0; i < nLoops * numCons; ++i) {
		int rNum = rand() % RAND_DIVISOR;
		sleep(rNum);
		item = rand() % 100;
		sem_wait(&empty);
		pthread_mutex_lock(&mutex);
		if(insert_item(item))
			fprintf(stderr, "Producer reports error condition\n");
		else
			printf("producer %2d produced %2d\n", num, item);
		pthread_mutex_unlock(&mutex);
		sem_post(&full);
	}
	return NULL;
}

void *consumer(void *param) {
	buffer_item item;
	int num = *((int*)param);
	for(int i = 0; i < nLoops * numProd; ++i) {
		int rNum = rand() % RAND_DIVISOR;
		sleep(rNum);
		sem_wait(&full);
		pthread_mutex_lock(&mutex);
		remove_item(&item);
		printf("consumer %2d consumed %2d\n", num, item);
		pthread_mutex_unlock(&mutex);
		sem_post(&empty);
	}
	return NULL;
}

int insert_item(buffer_item item) {
	buffer[counter] = item;
	counter++;
	return 0;
}

int remove_item(buffer_item *item) {
	counter--;
	*item = buffer[counter];
	return 0;
}

int main(int argc, char *argv[]) {
	if(argc != 4) {
		fprintf(stderr, "usage:%s <nLoops> <nProd> <nCons>\n", argv[0]);
		exit(0);
	}
	srand(time(0));
	nLoops = atoi(argv[1]);		//	Number of loops per thread
	numProd = atoi(argv[2]);	//	Number of producer threads
	numCons = atoi(argv[3]);	//	Number of consumer threads
//	nLoops *= numProd * numCons;
	int np[numProd];
	int nc[numCons];
	pthread_t tp[numProd];
	pthread_t tc[numCons];

	initializeData();
	printf("\n");
	for(int i = 0; i < numProd; i++) {
		np[i] = i+1;
		pthread_create(tp+i, 0, producer, np+i);
	}
	for(int i = 0; i < numCons; i++) {
		nc[i] = i+1;
		pthread_create(tc+i, 0, consumer, nc+i);
	}
	for(int i = 0; i < numProd; i++) pthread_join(tp[i], NULL);
	for(int i = 0; i < numCons; i++) pthread_join(tc[i], NULL);
	sem_destroy(&full);
	sem_destroy(&empty);
	printf("\n");
	printf("main() exited\n\n");
	exit(0);
}

