#define _XOPEN_SOURCE
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/wait.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

//	The semaphore key is an arbitrary long integer which serves as an external
//	identifier by which the semaphore is known to any program that wishes to use
//	it.

#define KEY (1492)

int main() {

//	Create the semaphore with external key KEY if it doesn't already exists.
//	Give permissions to the world.

	int id = semget(KEY, 1, 0666 | IPC_CREAT);
	if(id < 0) {
		fprintf(stderr, "Unable to obtain semaphore.\n");
		exit(0);
	}

//	What we actually get is an array of semaphores. The second argument to
//	semget() was the array dimension - in our case 1.

//	The next thing is an argument to the semctl() function. semctl() does
//	various things to the semaphore depending on which arguments are passed. We
//	will use it to make sure that the value of the semaphore is initially 0.

//	int semctl(int semid, int semnum, int cmd, ...);

//	semctl() performs the control operation specified by cmd on the semaphore
//	set identified by semid, or on the semnum-th semaphore of that set. (The
//	semaphores in a set are numbered starting at 0.)

//	This function has three or four arguments, depending on cmd. When there are
//	four, the fourth has the type union semun. The calling program must define
//	this union as follows: 

	union semun {
		int val;	//	Value for SETVAL
		struct semid_ds *buf;	//	Buffer for IPC_STAT, IPC_SET
		unsigned short *array;	//	Array for GETALL, SETALL
		struct seminfo *__buf;	//	Buffer for IPC_INFO
	} argument;

	argument.val = 0;
	if(semctl(id, 0, SETVAL, argument) < 0)
		fprintf( stderr, "Cannot set semaphore value.\n");
	else
		fprintf(stderr, "Semaphore %d initialized.\n", KEY);

	pid_t pidb = fork();
	if(pidb == 0) { // child, execute semb
		execl("semb", "semb", NULL);
	}
	sleep(2);
	pid_t pida = fork();
	if(pida == 0) { // child, execute semb
		execl("sema", "sema", NULL);
	}
	wait(0);
	wait(0);
	semctl(id, 0, IPC_RMID);
	printf("main exit\n\n");

	return 0;
}

